/*
 * snifscan -- packet sniffer scanner banner grabber
 *
 * skape
 * 06/01/2002
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <netinet/in_systm.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#define __FAVOR_BSD
#include <netinet/tcp.h>
#include <signal.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>

#ifdef IP
#undef IP
#endif

#include <pthread.h>
#include <pcap.h>

#include "ipblocks.h"

#define IP  struct ip
#define TCP struct tcphdr

typedef struct _tcpinfo {

	unsigned short srcPort;
	unsigned short dstPort;

	unsigned char flags;
	unsigned long seq;
	unsigned long ack;
	
	unsigned long optionsLength;
	unsigned char *options;

	unsigned long payloadLength;
	unsigned char *payload;
	unsigned char payloadNeedFree;

} TCPINFO;

typedef struct _global {

	FILE *logFd;

	char *interface;
	int rawFd;

	char *targetRange;
	unsigned long source;

	unsigned int lowPort;
	unsigned int highPort;

	int parentPid;
	int initCode;

	int delay;
	int rangeOfDelay;

	pthread_mutex_t initMutex;
	pthread_mutex_t fdMutex;

} GLOBAL;

/* A return value of 0 from a port handler informs snifscan that underlying functions
 * should not be used. 
 */
char port_handler_telnet_ack(TCPINFO *tcpinfo, unsigned long srcAddr, unsigned char *data, unsigned long dataLen);

char port_handler_httpd_synack(TCPINFO *tcpinfo, unsigned long srcAddr);

struct _port_handlers {
	
	unsigned short port;
	char (*synAckFunction)(TCPINFO *tcpinfo, unsigned long srcAddr);
	char (*ackFunction)(TCPINFO *tcpinfo, unsigned long srcAddr, unsigned char *data, unsigned long dataLen);

} port_handlers[] = {

	{ 23,		NULL,	port_handler_telnet_ack		},
	{ 80,		port_handler_httpd_synack,	NULL	},
	{ 0,		NULL,				NULL	},

};

GLOBAL global;

unsigned char initialTcpOptions[] = {
	0x02, 0x04, 0x05, 0xb4,				/* max segment size = 1460 */
	0x04, 0x02,								/* SACK permitted */
	0x08, 0x0a, 0x0f, 0x81, 0xB1, 0xB2, 0x00, 0x00, 0x00, 0x00, /* Time stamp */
	0x01,										/* NOP */
	0x03, 0x03, 0x00						/* Window scale: 0 */
};

unsigned long resolve(const char *addr);
unsigned short ipChecksum(unsigned short *ptr, int nbytes);
int sendTcp(unsigned long dstAddr, unsigned long srcAddr, TCPINFO *tcpinfo);
void watchPackets(void *null);
void executeScan();

int main(int argc, char **argv)
{
	extern char *optarg;
	int c, set = 1;
	pthread_t id;

	if (argc == 1)
	{
		fprintf(stdout,
					"Usage: %s [-t target range] [-s src addr] [-l low port] [-h high port]\n"
					"       [-i interface] [-d delay] [-r delay range] [-o output file]\n\n"
					" Target ranges supported syntax:\n"
					"    10.0.0/24\n"
					"    10.0.0.0/255.255.255.0\n"
					"    1.2.3.4\n\n"
					" Example usage: Scan 10.0.0.0/24 with the src addr of 1.2.3.4 for ports 21, 22, and 23\n"
					"     %s -t 10.0.0.0/24 -s 1.2.3.4 -l 21 -h 23 -i eth0 -o 10.0.0.out\n",argv[0],argv[0]);

		return 0;	
	}

	if (getuid() != 0)
	{
		fprintf(stdout,"You must be root to use this program.\n");
		return 0;
	}

	memset(&global, 0, sizeof(GLOBAL));

	srand(time(NULL) | getpid());

	global.targetRange = NULL;
	global.source = -1;
	global.interface = "eth0";
	global.delay = 1000;				/* 1000 ms */
	global.rangeOfDelay = 128;		/* 128 ips */
	global.logFd = stdout;
  
	pthread_mutex_init(&global.initMutex, NULL); 
	pthread_mutex_init(&global.fdMutex, NULL);
	
	while ((c = getopt(argc, argv, "t:s:l:h:i:d:r:o:")) != EOF)
	{
		switch (c)
		{
			case 't':
				global.targetRange = optarg;
				break;
			case 's':
				global.source = inet_addr(optarg);
				break;
			case 'l':
				global.lowPort = atoi(optarg);
				break;
			case 'h':
				global.highPort = atoi(optarg);
				break;
			case 'i':
				global.interface = optarg;
				break;	
			case 'd':
				global.delay = atoi(optarg);
				break;
			case 'r':
				global.rangeOfDelay = atoi(optarg);
				break;
			case 'o':
				if (!(global.logFd = fopen(optarg, "w")))
				{
					fprintf(stdout,"could not open logfile '%s'\n",optarg);
					return 0;
				}
				break;
		}
	}

	if ((!global.targetRange) || (global.source == -1) || (global.lowPort <= 0) || (global.highPort <= 0))
	{
		fprintf(stdout,"One or more arguments specified was invalid.\n");
		return 0;
	}

	if ((global.rawFd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) <= 0)
	{
		perror("socket");
		return 0;
	}

	if (setsockopt(global.rawFd, IPPROTO_IP, IP_HDRINCL, (char *) &set, sizeof(set)) < 0)
	{
		perror("setsockopt");
		return 0;
	}

	pthread_mutex_lock(&global.initMutex);

	pthread_create(&id, NULL, (void *(*)(void *))watchPackets, NULL);

	pthread_mutex_lock(&global.initMutex);

	if (global.initCode != 1)
		return (int)fprintf(stdout,"failed to start packet sniffer.\n");

	executeScan();

	pthread_mutex_lock(&global.initMutex);

	fprintf(stdout,"TCP handshake completed.\n");

	return 1;
}

unsigned long resolve(const char *addr)
{
	unsigned long ret;
	struct hostent *h;

	if ((ret = inet_addr(addr)) != -1) 
		return ret;

	if (!(h = gethostbyname(addr)))
		return -1;

	memcpy(&ret, h->h_addr, h->h_length);

	return ret;
}

unsigned short ipChecksum(unsigned short *ptr, int nbytes)
{
	register long sum;        
	register u_short answer; 
	u_short oddbyte;
   
	sum = 0;

	while (nbytes > 1)
	{   
		sum += *ptr++;
		nbytes -= 2;
	}
        
	if (nbytes == 1)
	{   
		oddbyte = 0;     
		*((u_char *) & oddbyte) = *(u_char *) ptr;     
		sum += oddbyte;
	}
   
	sum = (sum >> 16) + (sum & 0xffff); 
	sum += (sum >> 16);       
	answer = ~sum;           

	return (answer);
}

int sendTcp(unsigned long dstAddr, unsigned long srcAddr, TCPINFO *tcpinfo)
{
	unsigned int packetLen, minimalLength;
	unsigned char packet[(packetLen = sizeof(IP) + sizeof(TCP) + tcpinfo->optionsLength + tcpinfo->payloadLength)];
	IP *iphdr = (IP *)packet;
	TCP *tcphdr = (TCP *)((unsigned char *)packet + sizeof(IP));
	unsigned char *options = ((unsigned char *)packet + sizeof(IP) + sizeof(TCP));
	unsigned char *payload = ((unsigned char *)packet + sizeof(IP) + sizeof(TCP) + tcpinfo->optionsLength);
	struct sockaddr_in s;

	unsigned char minimalip[(minimalLength = 12 + sizeof(TCP) + tcpinfo->optionsLength + tcpinfo->payloadLength)];

	memset(&s, 0, sizeof(s));
	memset(packet, 0, packetLen);

	memset(minimalip, 0, minimalLength);
	*((unsigned long *)((unsigned char *)minimalip+0)) = srcAddr;
	*((unsigned long *)((unsigned char *)minimalip+4)) = dstAddr;
	*((unsigned char *)((unsigned char *)minimalip+8)) = 0;
	*((unsigned char *)((unsigned char *)minimalip+9)) = IPPROTO_TCP;
	*((unsigned short *)((unsigned char *)minimalip+10)) = htons(packetLen - sizeof(IP));

	iphdr->ip_v       = 4;
	iphdr->ip_hl      = 5;
	iphdr->ip_id      = rand() & 0xFFFF;
	iphdr->ip_src.s_addr = srcAddr;
	iphdr->ip_dst.s_addr = dstAddr;
	iphdr->ip_p       = IPPROTO_TCP;
	iphdr->ip_ttl     = 40;
	iphdr->ip_len     = packetLen;
	iphdr->ip_sum     = ipChecksum((u_short *)iphdr, sizeof(IP));

	s.sin_family      = AF_INET;
	s.sin_addr.s_addr = dstAddr;

	tcphdr->th_sport = htons(tcpinfo->srcPort);
	tcphdr->th_dport = htons(tcpinfo->dstPort);
	tcphdr->th_seq   = htonl(tcpinfo->seq);
	tcphdr->th_ack   = htonl(tcpinfo->ack);
	tcphdr->th_off   = (sizeof(TCP) / 4) + (tcpinfo->optionsLength / 4);
	tcphdr->th_flags = tcpinfo->flags;
	tcphdr->th_win   = htons(5840);

	if (tcpinfo->optionsLength > 0)
		memcpy(options, tcpinfo->options, tcpinfo->optionsLength);
	if (tcpinfo->payloadLength > 0)
		memcpy(payload, tcpinfo->payload, tcpinfo->payloadLength);

	memcpy(minimalip+12, ((unsigned char *)packet)+sizeof(IP), packetLen - sizeof(IP));

	tcphdr->th_sum   = ipChecksum((u_short *)&minimalip, minimalLength);

	pthread_mutex_lock(&global.fdMutex);

	sendto(global.rawFd, packet, packetLen, 0, (struct sockaddr *)&s, sizeof(s));

	pthread_mutex_unlock(&global.fdMutex);

	if (tcpinfo->payloadNeedFree)
		free(tcpinfo->payload);

	return 1;
}

void watchPackets(void *null)
{
	char errorBuffer[PCAP_ERRBUF_SIZE];
	struct pcap_pkthdr packetHeader;
	pcap_t *pcapFd;
	unsigned char *newPacket;
	IP *iphdr;
	int x = 0;

	global.initCode = 0;

	pthread_detach(pthread_self());

	if (!(pcapFd = pcap_open_live(global.interface, 128, 0, 0, errorBuffer)))
		fprintf(stdout,"watchPackets(): failed to open interface '%s'.\n",global.interface);
	else
		global.initCode = 1;
	
	pthread_mutex_unlock(&global.initMutex);

	if (!pcapFd)
		pthread_exit(0);

	while ((newPacket = (char *)pcap_next(pcapFd, &packetHeader)) != NULL)
	{
		/* Make sure packet is a valid TCP packet length first. */
		if ((!newPacket) || (packetHeader.len < (16 + sizeof(IP) + sizeof(TCP))))
			continue;

		iphdr = (IP *)((unsigned char *)(newPacket) + 14);

		if (iphdr->ip_v != 0x4)
			continue;

		if ((iphdr->ip_dst.s_addr == global.source) &&
			(iphdr->ip_p == IPPROTO_TCP))
		{
			TCP *tcphdr = (TCP *)((unsigned char *)(iphdr) + sizeof(IP));
			unsigned char *data = ((unsigned char *)iphdr) + sizeof(IP) + (4 * tcphdr->th_off);
			unsigned short dataSize = ntohs(iphdr->ip_len) - ((4 * iphdr->ip_hl) + (4 * tcphdr->th_off));
			unsigned char *options = NULL;
			TCPINFO tcpinfo = {
					srcPort: ntohs(tcphdr->th_dport),
					dstPort: ntohs(tcphdr->th_sport),
					flags:   TH_ACK,
					seq:     ntohl(tcphdr->th_ack),
					ack:     ntohl(tcphdr->th_seq),
					optionsLength: (tcphdr->th_off - (sizeof(TCP)/4)) * 4,
					options:       ((unsigned char *)tcphdr)+(sizeof(TCP)),
					payloadLength: 0,
					payload: NULL,
					payloadNeedFree: 0
				};
			unsigned char ret = 1;

			if (dataSize)
				tcpinfo.ack += dataSize;
			else
				tcpinfo.ack++;

			if (tcphdr->th_off > 5)
			{
				/* Some options must be copied. */
				unsigned long optionsLength = 0, incomingLength = (tcphdr->th_off-5) * 4, optionsPosition;
				unsigned char *currentOptions = ((unsigned char *)tcphdr) + sizeof(TCP);
				unsigned long t = time(NULL);

				x = 0;

				while (x < incomingLength)
				{
					switch (currentOptions[x])
					{
						case 0:		/* 0 pad */
						case 1:		/* NOP   */
							x++;
							break;
						case 2:		/* MSS   */
							x+=4;
							break;
						case 3:		/* Window scale option */
							x+=3;
							break;
						case 8:		/* Timestamp */
							optionsPosition = optionsLength;
							if (!options)
								options = (char *)malloc((optionsLength = 12));
							else
								options = (char *)realloc(options, (optionsLength += 12));

							options[optionsPosition++] = 0x01;
							options[optionsPosition++] = 0x01;
							options[optionsPosition++] = 0x08;
							options[optionsPosition++] = 0x0a;

							memcpy(options+optionsPosition,&t,4);
							optionsPosition += 4;

							if (x+10 <= incomingLength)
								memcpy(options+optionsPosition,&currentOptions[x+2],4);

							x+=10;
							break;
						case 4:		/* SACK permitted */
							x+=2;
							break;
						case 5:		/* sack option */	
							if ((x+1 < incomingLength))
								x+=currentOptions[x+1]+1;
							else
								x++;
							break;
					}
				}

				tcpinfo.optionsLength = optionsLength;
				tcpinfo.options = options;
			}

			if ((tcphdr->th_flags & TH_ACK) && (tcphdr->th_flags & TH_SYN))
			{
				fprintf(global.logFd,"%hi.%hi.%hi.%hi: Port open %d.\n", 
								((unsigned char *)&iphdr->ip_src.s_addr)[0],
								((unsigned char *)&iphdr->ip_src.s_addr)[1],
								((unsigned char *)&iphdr->ip_src.s_addr)[2],
								((unsigned char *)&iphdr->ip_src.s_addr)[3],
								ntohs(tcphdr->th_sport));

				for (x = 0; port_handlers[x].port; x++)
				{
					if ((port_handlers[x].port == tcpinfo.dstPort) && (port_handlers[x].synAckFunction))
					{
						ret = port_handlers[x].synAckFunction(&tcpinfo, iphdr->ip_src.s_addr);
						break;
					}
				}

				if (!ret)
					continue;	

				sendTcp(iphdr->ip_src.s_addr, iphdr->ip_dst.s_addr, &tcpinfo);
			}
			else
			{
				if (!(tcphdr->th_flags & TH_RST) && !(tcphdr->th_flags & TH_FIN))
				{
					for (x = 0; port_handlers[x].port; x++)
					{
						if ((port_handlers[x].port == tcpinfo.dstPort) && (port_handlers[x].ackFunction))
						{
							ret = port_handlers[x].ackFunction(&tcpinfo, iphdr->ip_src.s_addr, data, dataSize);
							break;
						}
					}	

					if (!dataSize || !ret)
					{
						if (options)
							free(options);

						continue;
					}

					if (ret != 2)
					{
						fprintf(global.logFd,"%hi.%hi.%hi.%hi: Data on port %d (size=%d): ",
									((unsigned char *)&iphdr->ip_src.s_addr)[0],
									((unsigned char *)&iphdr->ip_src.s_addr)[1],
									((unsigned char *)&iphdr->ip_src.s_addr)[2],
									((unsigned char *)&iphdr->ip_src.s_addr)[3],
									ntohs(tcphdr->th_sport),
									dataSize);
	
						for (x = 0; x < dataSize; x++)
							fprintf(stdout,"%c",((data[x] >= 32) && (data[x] <= 127))?data[x]:'.');
	
						fprintf(stdout,"\n");
					}

					sendTcp(iphdr->ip_src.s_addr, iphdr->ip_dst.s_addr, &tcpinfo);

				}
			}
					
			if (options)
				free(options);
		}
	}

	pthread_mutex_unlock(&global.initMutex);
}

void executeScan()
{
	unsigned long ips[global.rangeOfDelay], x = 0, currentPort;
	IPBLOCK *ipblock = ipblocksCreateIpBlock(global.targetRange,32,0);
	TCPINFO tcpinfo;
	time_t t = time(NULL);

	if (!ipblock)
	{
		fprintf(stdout,"The string specified was not valid. '%s'\n",global.targetRange);
		return;
	}

	memset(&tcpinfo, 0, sizeof(tcpinfo));
	tcpinfo.flags = TH_SYN;
	tcpinfo.optionsLength = sizeof(initialTcpOptions);
	tcpinfo.options = initialTcpOptions;

	memcpy(initialTcpOptions+8,&t,sizeof(t)); 

	while (ipblocksEnumerate(ipblock,ips,global.rangeOfDelay))
	{
		for (x = 0; x < 128 && ips[x]; x++)
		{
			if (((((unsigned char *)(&ips[x]))[3]) == 255) || ((((unsigned char *)(&ips[x]))[3]) == 0))
				continue;

			for (currentPort = global.lowPort; currentPort <= global.highPort; currentPort++)
			{
				tcpinfo.srcPort = (rand() & 0xFFFE) + 1;
				tcpinfo.dstPort = currentPort;
				tcpinfo.seq     = rand();
				tcpinfo.ack     = rand();

				sendTcp(ips[x], global.source, &tcpinfo);
			}
		}

		usleep(global.delay);
	}	

	ipblocksDestroyIpBlock(ipblock);
}

/* Port handlers */

char port_handler_telnet_ack(TCPINFO *tcpinfo, unsigned long srcAddr, unsigned char *data, unsigned long dataLen)
{
	unsigned long payloadPosition;
	int x = 0;

	if (dataLen <= 0)
		return 1;

	fprintf(global.logFd,"%hi.%hi.%hi.%hi: Data on port %d: ",
					((unsigned char *)&srcAddr)[0],
					((unsigned char *)&srcAddr)[1],
					((unsigned char *)&srcAddr)[2],
					((unsigned char *)&srcAddr)[3],
					tcpinfo->dstPort);
	fflush(global.logFd);

	while (x < dataLen)
	{
		if (data[x] == IAC)
		{
			unsigned char cmd = data[++x];

			payloadPosition = tcpinfo->payloadLength;

			switch (cmd)
			{
				case DO:
				case WILL:
				default:
					if (tcpinfo->payload)
						tcpinfo->payload = (char *)realloc(tcpinfo->payload,(tcpinfo->payloadLength = tcpinfo->payloadLength + 3));
					else
						tcpinfo->payload = (char *)malloc((tcpinfo->payloadLength = 3));
					
					tcpinfo->payload[payloadPosition++] = IAC;	
					tcpinfo->payload[payloadPosition++] = (cmd == DO)?WONT:DONT;	
					tcpinfo->payload[payloadPosition++] = data[++x];	
					break;
			}	
		}
		else
			fprintf(global.logFd,"%c",((data[x] == 13) || (data[x] == 10) || ((data[x] >= 32) && (data[x] <= 127)))?data[x]:'.');

		x++;
	}

	fprintf(global.logFd,"\n");

	if (tcpinfo->payload)
		tcpinfo->payloadNeedFree = 1;
	
	return 2;
}

char port_handler_httpd_synack(TCPINFO *tcpinfo, unsigned long srcAddr)
{
	tcpinfo->payload = "GET / HTTP/1.0\r\n\r\n";
	tcpinfo->payloadLength = strlen(tcpinfo->payload);

	return 1;
}
